# -*- coding: utf-8 -*-

# Copyright (c) 2008/2013 Andrey Vlasovskikh
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


def pretty_tree(x, kids, show):
    """(a, (a -> list(a)), (a -> str)) -> str

    Returns a pseudographic tree representation of x similar to the tree command
    in Unix.
    """
    (MID, END, CONT, LAST, ROOT) = ('|-- ', '`-- ', '|   ', '    ', '')

    def rec(x, indent, sym):
        line = indent + sym + show(x)
        xs = kids(x)
        if len(xs) == 0:
            return line
        else:
            if sym == MID:
                next_indent = indent + CONT
            elif sym == ROOT:
                next_indent = indent + ROOT
            else:
                next_indent = indent + LAST
            syms = [MID] * (len(xs) - 1) + [END]
            lines = [rec(x, next_indent, sym) for x, sym in zip(xs, syms)]
            return '\n'.join([line] + lines)

    return rec(x, '', ROOT)



#!/usr/bin/env python3
'''
KLL Data Organization
'''

# Copyright (C) 2016-2018 by Jacob Alexander
#
# This file is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This file is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this file.  If not, see <http://www.gnu.org/licenses/>.

### Imports ###

import copy
import re

import kll.common.expression as expression

from kll.common.id import PixelAddressId


### Decorators ###

# Print Decorator Variables
ERROR = '\033[5;1;31mERROR\033[0m:'
WARNING = '\033[5;1;33mWARNING\033[0m:'

ansi_escape = re.compile(r'\x1b[^m]*m')


### Classes ###

class Data:
    '''
    Base class for KLL datastructures
    '''
    # Debug output formatters
    debug_output = {
        'add': "\t\033[1;42;37m++\033[0m\033[1mADD KEY\033[1;42;37m++\033[0m \033[1m<==\033[0m {0}",
        'app': "\t\033[1;45;37m**\033[0m\033[1mAPP KEY\033[1;45;37m**\033[0m \033[1m<==\033[0m {0}",
        'mod': "\t\033[1;44;37m##\033[0m\033[1mMOD KEY\033[1;44;37m##\033[0m \033[1m<==\033[0m {0}",
        'rem': "\t\033[1;41;37m--\033[0m\033[1mREM KEY\033[1;41;37m--\033[0m \033[1m<==\033[0m {0}",
        'drp': "\t\033[1;43;37m@@\033[0m\033[1mDRP KEY\033[1;43;37m@@\033[0m \033[1m<==\033[0m {0}",
        'dup': "\t\033[1;46;37m!!\033[0m\033[1mDUP KEY\033[1;46;37m!!\033[0m \033[1m<==\033[0m {0}",
    }

    def __init__(self, parent):
        '''
        Initialize datastructure

        @param parent: Parent organization, used to query data from other datastructures
        '''
        self.data = {}
        self.parent = parent
        self.connect_id = 0
        self.merge_in_log = []

    def merge_in_log_prune(self, debug):
        '''
        Prune the merge_in_log

        Reverse searches the list, if the key already exists, disable the key
        '''
        new_log = []
        found = []
        # We have to manually reverse, then modify the referenced items
        # i.e. we're still modifying self.merge_in_log
        # This is done so we still have a proper index, and do the invalidation in the correct order
        reversed_log = self.merge_in_log[::-1]
        for index, elem in enumerate(reversed_log):
            key, expr, enabled = elem
            # Add to found list
            if key not in found:
                found.append(key)
                new_log.insert(0, elem)
            # Otherwise mark as disabled
            else:
                reversed_log[index] = [
                    key,
                    expr,
                    False,
                ]

        return new_log

    def merge_in_log_expression(self, key, expression, debug):
        '''
        Logs a given merge_in expressions
        This is used to determine the order in which the merges occurred

        Duplicate entries are pruned after the merge

        @param key:        Hash entry for (text)
        @param expression: Expression object
        @param debug:      Enable debug out
        '''
        # Debug output
        if debug[0]:
            output = "{0} Log Add: {1} {2}".format(
                self.parent.parent.layer_info(),
                key,
                expression,
            )
            print(debug[1] and output or ansi_escape.sub('', output))

        # Add to log, and enable key
        self.merge_in_log.append([key, expression, True])

    def add_expression(self, expression, debug):
        '''
        Add expression to data structure

        May have multiple keys to add for a given expression

        @param expression: KLL Expression (fully tokenized and parsed)
        @param debug:      Enable debug output
        '''
        # Lookup unique keys for expression
        keys = expression.unique_keys()

        # Add/Modify expressions in datastructure
        for key, uniq_expr in keys:
            # Check which operation we are trying to do, add or modify
            if debug[0]:
                if key in self.data.keys():
                    output = self.debug_output['mod'].format(key)
                else:
                    output = self.debug_output['add'].format(key)
                print(debug[1] and output or ansi_escape.sub('', output))

            self.data[key] = uniq_expr

            # Add to log
            self.merge_in_log_expression(key, uniq_expr, debug)

    def merge(self, merge_in, map_type, debug):
        '''
        Merge in the given datastructure to this datastructure

        This datastructure serves as the base.

        @param merge_in: Data structure from another organization to merge into this one
        @param map_type: Used fo map specific merges
        @param debug:    Enable debug out
        '''
        # The default case is just to add the expression in directly
        for key, kll_expression in merge_in.data.items():
            # Set ConnectId in expression
            kll_expression.connect_id = merge_in.connect_id

            # Display key:expression being merged in
            if debug[0]:
                output = merge_in.elem_str(key, True)
                print(debug[1] and output or ansi_escape.sub('', output), end="")

            self.add_expression(kll_expression, debug)

            # Add to log
            self.merge_in_log_expression(key, kll_expression, debug)

    def reduction(self, debug=False):
        '''
        Simplifies datastructure

        Most of the datastructures don't have a reduction. Just do nothing in this case.
        '''
        pass

    def cleanup(self, debug=False):
        '''
        Post-processing step for merges that may need to remove some data in the organization.
        Mainly used for dropping BaseMapContext expressions after generating a PartialMapContext.
        '''
        pass

    def connectid(self, connect_id):
        '''
        Sets the Data store with a given connect_id
        By default, this is 0, but may be set prior to an organization merge
        '''
        self.connect_id = connect_id

    def elem_str(self, key, single=False):
        '''
        Debug output for a single element

        @param key:    Index to datastructure
        @param single: Setting to True will bold the key
        '''
        if single:
            return "\033[1;33m{0: <20}\033[0m \033[1;36;41m>\033[0m {1}\n".format(key, self.data[key])
        else:
            return "{0: <20} \033[1;36;41m>\033[0m {1}\n".format(key, self.data[key])

    def __repr__(self):
        output = ""

        # Display sorted list of keys, along with the internal value
        for key in sorted(self.data):
            output += self.elem_str(key)

        return output


class MappingData(Data):
    '''
    KLL datastructure for data mapping

    ScanCode  trigger -> result
    USBCode   trigger -> result
    Animation trigger -> result
    '''
    def add_expression(self, expression, debug):
        '''
        Add expression to data structure

        May have multiple keys to add for a given expression

        Map expressions insert into the datastructure according to their operator.

        +Operators+
        :   Add/Modify
        :+  Append
        :-  Remove
        ::  Lazy Add/Modify

        i:  Add/Modify
        i:+ Append
        i:- Remove
        i:: Lazy Add/Modify

        The i or isolation operators are stored separately from the main ones.
        Each key is pre-pended with an i

        The :: or lazy operators act just like : operators, except that they will be ignore if the evaluation
        merge cannot resolve a ScanCode.

        @param expression: KLL Expression (fully tokenized and parsed)
        @param debug:      Enable debug output
        '''
        # Lookup unique keys for expression
        keys = expression.unique_keys()

        # Add/Modify expressions in datastructure
        for ukey, uniq_expr in keys:
            # Determine which the expression operator
            operator = expression.operator

            # Except for the : operator, all others have delayed action
            # Meaning, they change behaviour depending on how Contexts are merged
            # This means we can't simplify yet
            # In addition, :+ and :- are stackable, which means each key has a list of expressions
            # We append the operator to differentiate between the different types of delayed operations
            key = "{0}{1}".format(operator, ukey)

            # Determine if key exists already
            exists = key in self.data.keys()

            # Add/Modify
            if operator in [':', '::', 'i:', 'i::']:
                debug_tag = exists and 'mod' or 'add'

            # Append/Remove
            else:
                # Check to make sure we haven't already appended expression
                # Use the string representation to do the comparison (general purpose)
                if exists and "{0}".format(uniq_expr) in ["{0}".format(elem) for elem in self.data[key]]:
                    debug_tag = 'dup'

                # Append
                elif operator in [':+', 'i:+']:
                    debug_tag = 'app'

                # Remove
                else:
                    debug_tag = 'rem'

            # Debug output
            if debug[0]:
                output = self.debug_output[debug_tag].format(key)
                print(debug[1] and output or ansi_escape.sub('', output))

            # Don't append if a duplicate
            if debug_tag == 'dup':
                continue

            # Append, rather than replace
            if operator in [':+', ':-', 'i:+', 'i:-']:
                if exists:
                    self.data[key].append(uniq_expr)

                # Create initial list
                else:
                    self.data[key] = [uniq_expr]
            else:
                self.data[key] = [uniq_expr]

            # Append to log
            self.merge_in_log_expression(key, uniq_expr, debug)

    def set_interconnect_id(self, interconnect_id, triggers):
        '''
        Traverses the sequence of combo of identifiers to set the interconnect_id
        '''
        for sequence in triggers:
            for combo in sequence:
                for identifier in combo:
                    identifier.interconnect_id = interconnect_id

    def connectid(self, connect_id):
        '''
        Sets the Data store with a given connect_id
        By default, this is 0, but may be set prior to an organization merge
        '''
        self.connect_id = connect_id

        # Update dictionary keys using new connect_id
        for key, value in self.data.items():
            if value[0].type == 'ScanCode':
                # Update connect_id, then regenerate dictionary item
                value[0].connect_id = connect_id
                new_key = "{0}{1}".format(
                    value[0].operator,
                    value[0].unique_keys()[0][0],
                )

                # Replace dictionary item
                self.data[new_key] = self.data.pop(key)

    def maxscancode(self):
        '''
        Find max scancode per connect id

        @return: Dictionary of max Scan Codes (keys are the connect id)
        '''
        max_dict = {}
        for key, value in self.data.items():
            connect_id = value[0].connect_id
            max_uid = value[0].max_trigger_uid()

            # Initial value
            if connect_id not in max_dict.keys():
                max_dict[connect_id] = 0

            # Update if necessary
            if max_dict[connect_id] < max_uid:
                max_dict[connect_id] = max_uid

        return max_dict

    def merge_lazy_operators(self, debug):
        '''
        Lazy Set :: is not applied as a Set : until after the merge_in_log has been pruned

        Intended to be called during reduction.
        '''
        # Build dictionary of single ScanCodes first
        result_code_lookup = {}
        for key, expr in self.data.items():
            if expr[0].elems()[0] == 1 and expr[0].triggers[0][0][0].type == 'ScanCode':
                result_code_lookup.setdefault(expr[0].result_str(), []).append(key)

        # Build list of lazy keys from log
        lazy_keys = {}
        for key, expr, enabled in reversed(self.merge_in_log):
            if key[0:2] == '::' or key[0:3] == 'i::':
                if key not in lazy_keys.keys():
                    # Debug info
                    if debug:
                        print("\033[1mLazy\033[0m", key, expr)

                    # Determine the target key from the expression
                    target_key = expr.trigger_str()
                    lazy_keys[target_key] = expr

                    # Check if we need to do a lazy replacement
                    if target_key in result_code_lookup.keys():
                        expr_keys = result_code_lookup[target_key]
                        for target_expr_key in expr_keys:
                            # Calculate new key
                            new_expr = self.data[target_expr_key][0]
                            new_key = "{0}{1}".format(
                                new_expr.operator,
                                new_expr.unique_keys()[0][0]
                            )

                            # Determine action based on the new_expr.operator
                            orig_expr = self.data[new_key][0]

                            if debug:
                                print("\t\033[1;32mREPLACE\033[0m {0} -> {1}\n\t{2} => {3}".format(
                                    target_expr_key,
                                    new_key,
                                    expr,
                                    new_expr
                                ))

                            # Do replacement
                            self.data[new_key] = [expression.MapExpression(
                                    orig_expr.triggers,
                                    orig_expr.operator,
                                    expr.results
                            )]
                            self.data[new_key][0].connect_id = orig_expr.connect_id

                            # Unset basemap on expression
                            self.data[new_key][0].base_map = False

    def merge(self, merge_in, map_type, debug):
        '''
        Merge in the given datastructure to this datastructure

        This datastructure serves as the base.

        Map expressions merge differently than insertions.

        +Operators+
        :   Add/Modify       - Replace
        :+  Append           - Add
        :-  Remove           - Remove
        ::  Lazy Add/Modify  - Replace if found, otherwise drop

        i:  Add/Modify       - Replace
        i:+ Append           - Add
        i:- Remove           - Remove
        i:: Lazy Add/Modify  - Replace if found, otherwise drop

        @param merge_in: Data structure from another organization to merge into this one
        @param map_type: Used fo map specific merges
        @param debug:    Enable debug out
        '''
        # Get unique list of ordered keys
        # We can't just query the keys directly from the as we need them in order of being added
        # In addition, we need a unique list of keys, where the most recently added is the most important
        cur_keys = []
        for key, expr, enabled in reversed(merge_in.merge_in_log):
            if key not in cur_keys:
                cur_keys.insert(0, key)

        # Lazy Set ::
        lazy_keys = [key for key in cur_keys if key[0:2] == '::' or key[0:3] == 'i::']
        cur_keys = list(set(cur_keys) - set(lazy_keys))

        # Append :+
        append_keys = [key for key in cur_keys if key[0:2] == ':+' or key[0:3] == 'i:+']
        cur_keys = list(set(cur_keys) - set(append_keys))

        # Remove :-
        remove_keys = [key for key in cur_keys if key[0:2] == ':-' or key[0:3] == 'i:-']
        cur_keys = list(set(cur_keys) - set(remove_keys))

        # Set :
        # Everything left is just a set
        set_keys = cur_keys

        # First process the :: (or lazy) operators
        # We need to read into this datastructure and apply those first
        # Otherwise we may get undesired behaviour
        for key in lazy_keys:
            # Display key:expression being merged in
            if debug[0]:
                output = merge_in.elem_str(key, True)
                print(debug[1] and output or ansi_escape.sub('', output), end="")

            # Construct target key
            # XXX (HaaTa) We now delay lazy operation application till reduction
            #target_key = key[0] == 'i' and "i{0}".format(key[2:]) or key[1:]
            target_key = key

            # Lazy expressions will be dropped later at reduction
            debug_tag = 'mod'

            # Debug output
            if debug[0]:
                output = self.debug_output[debug_tag].format(key)
                print(debug[1] and output or ansi_escape.sub('', output))

            # Only replace
            self.data[target_key] = merge_in.data[key]

            # Unset BaseMapContext tag if not a BaseMapContext
            if map_type != 'BaseMapContext':
                self.data[target_key][0].base_map = False

        # Then apply : assignment operators
        for key in set_keys:
            # Display key:expression being merged in
            if debug[0]:
                output = merge_in.elem_str(key, True)
                print(debug[1] and output or ansi_escape.sub('', output), end="")

            # Construct target key
            target_key = key

            # Indicate if add or modify
            if target_key in self.data.keys():
                debug_tag = 'mod'
            else:
                debug_tag = 'add'

            # Debug output
            if debug[0]:
                output = self.debug_output[debug_tag].format(key)
                print(debug[1] and output or ansi_escape.sub('', output))

            # Set into new datastructure regardless
            self.data[target_key] = merge_in.data[key]

            # Unset BaseMap flag if this is not a BaseMap merge
            if map_type != 'BaseMapContext':
                self.data[target_key][0].base_map = False

        # Now apply append operations
        for key in append_keys:
            # Display key:expression being merged in
            if debug[0]:
                output = merge_in.elem_str(key, True)
                print(debug[1] and output or ansi_escape.sub('', output), end="")

            # Construct target key
            # XXX (HaaTa) Might not be correct, but seems to work with the merge_in_log
            #target_key = key[0] == 'i' and "i:{0}".format(key[3:]) or ":{0}".format(key[2:])
            target_key = key

            # Alwyays appending
            debug_tag = 'app'

            # Debug output
            if debug[0]:
                output = self.debug_output[debug_tag].format(key)
                print(debug[1] and output or ansi_escape.sub('', output))

            # Extend list if it exists
            if target_key in self.data.keys():
                self.data[target_key].extend(merge_in.data[key])
            else:
                self.data[target_key] = merge_in.data[key]

        # Finally apply removal operations to this datastructure
        # If the target removal doesn't exist, ignore silently (show debug message)
        for key in remove_keys:
            # Display key:expression being merged in
            if debug[0]:
                output = merge_in.elem_str(key, True)
                print(debug[1] and output or ansi_escape.sub('', output), end="")

            # Construct target key
            # XXX (HaaTa) Might not be correct, but seems to work with the merge_in_log
            #target_key = key[0] == 'i' and "i:{0}".format(key[3:]) or ":{0}".format(key[2:])
            target_key = key

            # Drop right away if target datastructure doesn't have target key
            if target_key not in self.data.keys():
                debug_tag = 'drp'

                # Debug output
                if debug[0]:
                    output = self.debug_output[debug_tag].format(key)
                    print(debug[1] and output or ansi_escape.sub('', output))

                continue

            # Compare expressions to be removed with the current set
            # Use strings to compare
            remove_expressions = ["{0}".format(expr) for expr in merge_in.data[key]]
            current_expressions = [("{0}".format(expr), expr) for expr in self.data[target_key]]
            for string, expr in current_expressions:
                debug_tag = 'drp'

                # Check if an expression matches
                if string in remove_expressions:
                    debug_tag = 'rem'

                # Debug output
                if debug[0]:
                    output = self.debug_output[debug_tag].format(key)
                    print(debug[1] and output or ansi_escape.sub('', output))

                # Remove if found
                if debug_tag == 'rem':
                    self.data[target_key] = [value for value in self.data.values() if value != expr]

        # Now append the merge_in_log
        self.merge_in_log.extend(merge_in.merge_in_log)

    def cleanup(self, debug=False):
        '''
        Post-processing step for merges that may need to remove some data in the organization.
        Mainly used for dropping BaseMapContext expressions after generating a PartialMapContext.
        '''
        # Using this dictionary, replace all the trigger USB codes
        # Iterate over a copy so we can modify the dictionary in place
        for key, expr in self.data.copy().items():
            if expr[0].base_map:
                if debug[0]:
                    output = "\t\033[1;34mDROP\033[0m {0}".format(self.data[key][0])
                    print(debug[1] and output or ansi_escape.sub('', output))
                del self.data[key]
            elif debug[0]:
                output = "\t\033[1;32mKEEP\033[0m {0}".format(self.data[key][0])
                print(debug[1] and output or ansi_escape.sub('', output))

    def reduction(self, debug=False):
        '''
        Simplifies datastructure

        Used to replace all trigger HIDCode(USBCode)s with ScanCodes

        NOTE: Make sure to create a new MergeContext before calling this as you lose data and prior context
        '''
        result_code_lookup = {}

        # Prune merge_in_log
        merge_in_pruned = self.merge_in_log_prune(debug)

        # Build dictionary of single ScanCodes first
        for key, expr in self.data.items():
            if expr[0].elems()[0] == 1 and expr[0].triggers[0][0][0].type == 'ScanCode':
                result_code_lookup[expr[0].result_str()] = expr

        # Skip if dict is empty
        if len(self.data.keys()) == 0:
            return

        # Instead of using the .data dictionary, use the merge_in_log which maintains the expression application order
        # Using this list, replace all the trigger USB codes
        for key, log_expr, active in self.merge_in_log:
            # Skip if not active
            if not active:
                continue

            # Lookup currently merged expression
            if key not in self.data.keys():
                continue
            expr = self.data[key]

            for sub_expr in expr:
                # 1) Single USB Codes trigger results will replace the original ScanCode result
                if sub_expr.elems()[0] == 1 and sub_expr.triggers[0][0][0].type in ['USBCode', 'SysCode', 'ConsCode']:
                    # Debug info
                    if debug:
                        print("\033[1mSingle\033[0m", key, expr)

                    # Lookup trigger to see if it exists
                    trigger_str = sub_expr.trigger_str()
                    if trigger_str in result_code_lookup.keys():
                        # Calculate new key
                        new_expr = result_code_lookup[trigger_str][0]
                        new_key = "{0}{1}".format(
                            new_expr.operator,
                            new_expr.unique_keys()[0][0]
                        )

                        # Determine action based on the new_expr.operator
                        orig_expr = self.data[new_key][0]

                        # Replace expression
                        if sub_expr.operator in [':']:
                            if debug:
                                print("\t\033[1;32mREPLACE\033[0m {0} -> {1}\n\t{2} => {3}".format(
                                    key,
                                    new_key,
                                    sub_expr,
                                    new_expr
                                ))

                            # Do replacement
                            self.data[new_key] = [expression.MapExpression(
                                    orig_expr.triggers,
                                    orig_expr.operator,
                                    sub_expr.results
                            )]

                            # Transfer connect_id
                            self.data[new_key][0].connect_id = orig_expr.connect_id

                            # Unset basemap on expression
                            self.data[new_key][0].base_map = False

                        # Add expression
                        elif sub_expr.operator in [':+']:
                            if debug:
                                print("\t\033[1;42mADD\033[0m {0} -> {1}\n\t{2} => {3}".format(
                                    key,
                                    new_key,
                                    sub_expr,
                                    new_expr
                                ))

                            # Add expression
                            self.data[new_key].append(expression.MapExpression(
                                orig_expr.triggers,
                                orig_expr.operator,
                                sub_expr.results
                            ))

                            # Unset basemap on sub results
                            for sub_expr in self.data[new_key]:
                                sub_expr.base_map = False

                        # Remove expression
                        elif sub_expr.operator in [':-']:
                            if debug:
                                print("\t\033[1;41mREMOVE\033[0m {0} -> {1}\n\t{2} => {3}".format(
                                    key,
                                    new_key,
                                    sub_expr,
                                    new_expr
                                ))

                        # Remove old key
                        if key in self.data.keys():
                            del self.data[key]

                    # Otherwise drop HID expression
                    else:
                        if debug:
                            print("\t\033[1;34mDROP\033[0m")
                        if key in self.data.keys():
                            del self.data[key]

                # 2) Complex triggers are processed to replace out any USB Codes with Scan Codes
                elif sub_expr.elems()[0] > 1:
                    # Debug info
                    if debug:
                        print("\033[1;4mMulti\033[0m ", key, expr)

                    # Lookup each trigger element and replace
                    # If any trigger element doesn't exist, drop expression
                    # Dive through sequence->combo->identifier (sequence of combos of ids)
                    replace = False
                    drop = False
                    for seq_in, sequence in enumerate(sub_expr.triggers):
                        for com_in, combo in enumerate(sequence):
                            for ident_in, identifier in enumerate(combo):
                                ident_str = "({0})".format(identifier)

                                # Replace identifier
                                if ident_str in result_code_lookup.keys():
                                    match_expr = result_code_lookup[ident_str]
                                    sub_expr.triggers[seq_in][com_in][ident_in] = match_expr[0].triggers[0][0][0]
                                    replace = True

                                # Ignore non-USB triggers
                                elif identifier.type in ['IndCode', 'GenericTrigger', 'Layer', 'LayerLock', 'LayerShift', 'LayerLatch', 'ScanCode']:
                                    pass

                                # Drop everything else
                                else:
                                    drop = True

                    # Trigger Identifier was replaced
                    if replace:
                        if debug:
                            print("\t\033[1;32mREPLACE\033[0m", expr)

                    # Trigger Identifier failed (may still occur if there was a replacement)
                    if drop:
                        if debug:
                            print("\t\033[1;34mDROP\033[0m")
                        del self.data[key]

        # Finally we can merge in the Lazy :: Set operators
        self.merge_lazy_operators(debug)

        # Show results of reduction
        if debug:
            print(self)


class AnimationData(Data):
    '''
    KLL datastructure for Animation configuration

    Animation -> modifiers
    '''


class AnimationFrameData(Data):
    '''
    KLL datastructure for Animation Frame configuration

    Animation -> Pixel Settings
    '''


class CapabilityData(Data):
    '''
    KLL datastructure for Capability mapping

    Capability -> C Function/Identifier
    '''


class DefineData(Data):
    '''
    KLL datastructure for Define mapping

    Variable -> C Define/Identifier
    '''


class PixelChannelData(Data):
    '''
    KLL datastructure for Pixel Channel mapping

    Pixel -> Channels
    '''

    def maxpixelid(self):
        '''
        Find max pixel id per connect id

        @return: dictionary of connect id to max pixel id
        '''
        max_pixel = {}
        for key, value in self.data.items():
            connect_id = value.connect_id

            # Make sure this is a PixelAddressId
            if isinstance(value.pixel.uid, PixelAddressId):
                max_uid = value.pixel.uid.index
            else:
                max_uid = value.pixel.uid

            # Initial value
            if connect_id not in max_pixel.keys():
                max_pixel[connect_id] = 0

            # Update if necessary
            if max_pixel[connect_id] < max_uid:
                max_pixel[connect_id] = max_uid

            # TODO REMOVEME
            #print( key,value, value.__class__, value.pixel.uid.index, value.connect_id )
        return max_pixel


class PixelPositionData(Data):
    '''
    KLL datastructure for Pixel Position mapping

    Pixel -> Physical Location
    '''

    def add_expression(self, expression, debug):
        '''
        Add expression to data structure

        May have multiple keys to add for a given expression

        @param expression: KLL Expression (fully tokenized and parsed)
        @param debug:      Enable debug output
        '''
        # Lookup unique keys for expression
        keys = expression.unique_keys()

        # Add/Modify expressions in datastructure
        for key, uniq_expr in keys:
            # Check which operation we are trying to do, add or modify
            if debug[0]:
                if key in self.data.keys():
                    output = self.debug_output['mod'].format(key)
                else:
                    output = self.debug_output['add'].format(key)
                print(debug[1] and output or ansi_escape.sub('', output))

            # If key already exists, just update
            if key in self.data.keys():
                self.data[key].update(uniq_expr)
            else:
                self.data[key] = uniq_expr

            # Append to log
            self.merge_in_log_expression(key, uniq_expr, debug)


class ScanCodePositionData(Data):
    '''
    KLL datastructure for ScanCode Position mapping

    ScanCode -> Physical Location
    '''

    def add_expression(self, expression, debug):
        '''
        Add expression to data structure

        May have multiple keys to add for a given expression

        @param expression: KLL Expression (fully tokenized and parsed)
        @param debug:      Enable debug output
        '''
        # Lookup unique keys for expression
        keys = expression.unique_keys()

        # Add/Modify expressions in datastructure
        for key, uniq_expr in keys:
            # Check which operation we are trying to do, add or modify
            if debug[0]:
                if key in self.data.keys():
                    output = self.debug_output['mod'].format(key)
                else:
                    output = self.debug_output['add'].format(key)
                print(debug[1] and output or ansi_escape.sub('', output))

            # If key already exists, just update
            if key in self.data.keys():
                self.data[key].update(uniq_expr)
            else:
                self.data[key] = uniq_expr

            # Append to log
            self.merge_in_log_expression(key, uniq_expr, debug)


class VariableData(Data):
    '''
    KLL datastructure for Variables and Arrays

    Variable -> Data
    Array    -> Data
    '''

    def add_expression(self, expression, debug):
        '''
        Add expression to data structure

        May have multiple keys to add for a given expression

        In the case of indexed variables, only replaced the specified index

        @param expression: KLL Expression (fully tokenized and parsed)
        @param debug:      Enable debug output
        '''
        # Lookup unique keys for expression
        keys = expression.unique_keys()

        # Add/Modify expressions in datastructure
        for key, uniq_expr in keys:
            # Check which operation we are trying to do, add or modify
            if debug[0]:
                if key in self.data.keys():
                    output = self.debug_output['mod'].format(key)
                else:
                    output = self.debug_output['add'].format(key)
                print(debug[1] and output or ansi_escape.sub('', output))

            # Check to see if we need to cap-off the array (a position parameter is given)
            if uniq_expr.type == 'Array' and uniq_expr.pos is not None:
                # Modify existing array
                if key in self.data.keys():
                    self.data[key].merge_array(uniq_expr)

                # Add new array
                else:
                    uniq_expr.merge_array()
                    self.data[key] = uniq_expr

            # Otherwise just add/replace expression
            else:
                self.data[key] = uniq_expr

            # Append to log
            self.merge_in_log_expression(key, uniq_expr, debug)


class Organization:
    '''
    Container class for KLL datastructures

    The purpose of these datastructures is to symbolically store at first, and slowly solve/deduplicate expressions.
    Since the order in which the merges occurs matters, this involves a number of intermediate steps.
    '''

    def __init__(self, parent):
        '''
        Intialize data structure
        '''
        self.parent = parent

        # Setup each of the internal sub-datastructures
        self.animation_data = AnimationData(self)
        self.animation_frame_data = AnimationFrameData(self)
        self.capability_data = CapabilityData(self)
        self.define_data = DefineData(self)
        self.mapping_data = MappingData(self)
        self.pixel_channel_data = PixelChannelData(self)
        self.pixel_position_data = PixelPositionData(self)
        self.scan_code_position_data = ScanCodePositionData(self)
        self.variable_data = VariableData(self)

        # Expression to Datastructure mapping
        self.data_mapping = {
            'AssignmentExpression': {
                'Array': self.variable_data,
                'Variable': self.variable_data,
            },
            'DataAssociationExpression': {
                'Animation': self.animation_data,
                'AnimationFrame': self.animation_frame_data,
                'PixelPosition': self.pixel_position_data,
                'ScanCodePosition': self.scan_code_position_data,
            },
            'MapExpression': {
                'TriggerCode': self.mapping_data,
                'PixelChannel': self.pixel_channel_data,
            },
            'NameAssociationExpression': {
                'Capability': self.capability_data,
                'Define': self.define_data,
            },
        }

    def __copy__(self):
        '''
        On organization copy, return a safe object

        Attempts to only copy the datastructures that may need to diverge
        '''
        new_obj = Organization(self.parent)

        # Copy only .data from each organization
        new_obj.animation_data.data = copy.copy(self.animation_data.data)
        new_obj.animation_frame_data.data = copy.copy(self.animation_frame_data.data)
        new_obj.capability_data.data = copy.copy(self.capability_data.data)
        new_obj.define_data.data = copy.copy(self.define_data.data)
        new_obj.mapping_data.data = copy.copy(self.mapping_data.data)
        new_obj.pixel_channel_data.data = copy.copy(self.pixel_channel_data.data)
        new_obj.pixel_position_data.data = copy.copy(self.pixel_position_data.data)
        new_obj.scan_code_position_data.data = copy.copy(self.scan_code_position_data.data)
        new_obj.variable_data.data = copy.copy(self.variable_data.data)

        # Also copy merge_in_log
        new_obj.animation_data.merge_in_log = copy.copy(self.animation_data.merge_in_log)
        new_obj.animation_frame_data.merge_in_log = copy.copy(self.animation_frame_data.merge_in_log)
        new_obj.capability_data.merge_in_log = copy.copy(self.capability_data.merge_in_log)
        new_obj.define_data.merge_in_log = copy.copy(self.define_data.merge_in_log)
        new_obj.mapping_data.merge_in_log = copy.copy(self.mapping_data.merge_in_log)
        new_obj.pixel_channel_data.merge_in_log = copy.copy(self.pixel_channel_data.merge_in_log)
        new_obj.pixel_position_data.merge_in_log = copy.copy(self.pixel_position_data.merge_in_log)
        new_obj.scan_code_position_data.merge_in_log = copy.copy(self.scan_code_position_data.merge_in_log)
        new_obj.variable_data.merge_in_log = copy.copy(self.variable_data.merge_in_log)

        return new_obj

    def stores(self):
        '''
        Returns list of sub-datastructures
        '''
        return [
            self.animation_data,
            self.animation_frame_data,
            self.capability_data,
            self.define_data,
            self.mapping_data,
            self.pixel_channel_data,
            self.pixel_position_data,
            self.scan_code_position_data,
            self.variable_data,
        ]

    def add_expression(self, expression, debug):
        '''
        Add expression to datastructure

        Will automatically determine which type of expression and place in the relevant store

        @param expression: KLL Expression (fully tokenized and parsed)
        @param debug:      Enable debug output
        '''
        # Determine type of of Expression
        expression_type = expression.__class__.__name__

        # Determine Expression Subtype
        expression_subtype = expression.type

        # Locate datastructure
        data = self.data_mapping[expression_type][expression_subtype]

        # Debug output
        if debug[0]:
            output = "\t\033[4m{0}\033[0m".format(data.__class__.__name__)
            print(debug[1] and output or ansi_escape.sub('', output))

        # Add expression to determined datastructure
        data.add_expression(expression, debug)

    def merge(self, merge_in, map_type, debug):
        '''
        Merge in the given organization to this organization

        This organization serves as the base.

        @param merge_in: Organization to merge into this one
        @param map_type: Used fo map specific merges
        @param debug:    Enable debug out
        '''
        # Merge each of the sub-datastructures
        for this, that in zip(self.stores(), merge_in.stores()):
            this.merge(that, map_type, debug)

    def cleanup(self, debug=False):
        '''
        Post-processing step for merges that may need to remove some data in the organization.
        Mainly used for dropping BaseMapContext expressions after generating a PartialMapContext.
        '''
        for store in self.stores():
            store.cleanup(debug)

    def reduction(self, debug=False):
        '''
        Simplifies datastructure

        NOTE: This will remove data, therefore, context is lost
        '''
        for store in self.stores():
            store.reduction(debug)

    def maxscancode(self):
        '''
        Find max scancode per connect id

        @return: dictionary of connect id to max scancode
        '''
        return self.mapping_data.maxscancode()

    def maxpixelid(self):
        '''
        Find max pixel id per connect id

        @return: dictionary of connect id to max pixel id
        '''
        return self.pixel_channel_data.maxpixelid()

    def __repr__(self):
        return "{0}".format(self.stores())



'''
regen test
Runs kll compiler in kll emitter mode, generating a kll file from a kll file.
After creating new kll files (for each stage), compare the final kll file with a known good one.
'''

### Imports ##

import difflib
import filecmp
import os
import pytest
import tempfile
from tests.klltest import kll_run, header_test



### Variables ###

test_files = [
    'kll/examples/assignment.kll',
    'kll/examples/capabilitiesExample.kll',
    'kll/examples/colemak.kll',
    'kll/examples/defaultMapExample.kll',
    'kll/examples/example.kll',
    'kll/examples/hhkbpro2.kll',
    'kll/examples/leds.kll',
    'kll/examples/mapping.kll',
    'kll/examples/md1Map.kll',
    'kll/examples/simple1.kll',
    'kll/examples/simple2.kll',
    'kll/examples/simpleExample.kll',
    'kll/examples/state_scheduling.kll',
    'kll/layouts/mouseTest.kll',
    'kll/layouts/klltest.kll',
    'kll/examples/triggers.kll',
]



### Tests ###

@pytest.mark.parametrize('input_file', test_files)
def test_regen(input_file):
    '''
    Runs regen test on each of the specified files
    '''
    # Prepare tmp directory
    sanitized_input_file = input_file.replace('/', '-')
    tmp_dir = os.path.join(tempfile.gettempdir(), 'kll_pytest')
    os.makedirs(tmp_dir, exist_ok=True)
    target_dir = tempfile.mkdtemp(suffix='-{}'.format(sanitized_input_file), prefix='regen-', dir=tmp_dir)

    # Determine cmp and new files
    cmp_filename = '{}_final.kll'.format(os.path.splitext(os.path.basename(input_file))[0])
    cmp_file = os.path.join('tests', 'cmp_regen', cmp_filename)
    new_file = os.path.join(target_dir, 'final.kll')

    # Run test
    args = ['--emitter', 'kll', '--output-debug', input_file, '--target-dir', target_dir]
    header_test('{} {} {}'.format(input_file, target_dir, cmp_file), args)
    ret = kll_run(args)
    assert ret == 0

    # Check if files are different
    if not filecmp.cmp(cmp_file, new_file):
        # Run diff as both files are different
        with open(cmp_file, 'r') as cmpfile:
            with open(new_file, 'r') as newfile:
                # Run diff and fail the test
                diff = difflib.ndiff(cmpfile.readlines(), newfile.readlines())
                print(''.join(diff), end="")
                assert False




'''
sanity test
Runs basic options to make sure Python can initialize and arguments can be used
'''

### Imports ###

from tests.klltest import kll_run



### Tests ###

def test_version():
    '''
    Calls --version argument
    '''
    args = ['--version']
    ret = kll_run(args)
    assert ret == 0

def test_help():
    '''
    Calls --help argument
    '''
    args = ['--help']
    ret = kll_run(args)
    assert ret == 0

def test_path():
    '''
    Calls --path argument
    '''
    args = ['--path']
    ret = kll_run(args)
    assert ret == 0

def test_layout_cache_path():
    '''
    Calls --layout-cache-path argument
    '''
    args = ['--layout-cache-path']
    ret = kll_run(args)
    assert ret == 0




'''
syntax test
Runs various .kll files through the compiler to make sure that they are processed without unexpected errors
'''

### Imports ##

import pytest
from tests.klltest import kll_run, header_test



### Variables ###

test_files = [
    'kll/examples/assignment.kll',
    'kll/examples/capabilitiesExample.kll',
    'kll/examples/colemak.kll',
    'kll/examples/defaultMapExample.kll',
    'kll/examples/example.kll',
    'kll/examples/hhkbpro2.kll',
    'kll/examples/leds.kll',
    'kll/examples/mapping.kll',
    'kll/examples/md1Map.kll',
    'kll/examples/simple1.kll',
    'kll/examples/simple2.kll',
    'kll/examples/simpleExample.kll',
    'kll/examples/state_scheduling.kll',
    'kll/layouts/mouseTest.kll',
    'kll/layouts/klltest.kll',
    'kll/examples/triggers.kll',
]



### Tests ###

@pytest.mark.parametrize('input_file', test_files)
def test_syntax(input_file):
    '''
    Runs syntax test on each of the specified files
    '''
    args = ['--emitter', 'none', '--data-finalization-display', input_file]
    header_test(input_file, args)
    ret = kll_run(args)
    assert ret == 0

@pytest.mark.parametrize('input_file', test_files)
def test_syntax_token_debug(input_file):
    '''
    Runs syntax test on each of the specified files, including --parser-token-debug
    '''
    args = ['--emitter', 'none', '--data-finalization-display', '--parser-token-debug', input_file]
    header_test(input_file, args)
    ret = kll_run(args)
    assert ret == 0

@pytest.mark.parametrize('input_file', test_files)
def test_syntax_operation_debug(input_file):
    '''
    Runs syntax test on each of the specified files, including --operation-organization-display
    '''
    args = ['--emitter', 'none', '--data-finalization-display', '--operation-organization-display', input_file]
    header_test(input_file, args)
    ret = kll_run(args)
    assert ret == 0

@pytest.mark.parametrize('input_file', test_files)
def test_syntax_fail_debug(input_file):
    '''
    Runs syntax test on each of the specified files, including failure debug options
    '''
    args = ['--emitter', 'none', '--token-debug', '--parser-token-debug', '--operation-organization-display', '--data-organization-display', '--data-finalization-display', input_file]
    header_test(input_file, args)
    ret = kll_run(args)
    assert ret == 0




#!/usr/bin/env python3
'''
KLL Expression Container
'''

# Copyright (C) 2016-2018 by Jacob Alexander
#
# This file is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This file is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this file.  If not, see <http://www.gnu.org/licenses/>.

### Imports ###

import copy

from kll.common.id import CapId



### Decorators ###

# Print Decorator Variables
ERROR = '\033[5;1;31mERROR\033[0m:'
WARNING = '\033[5;1;33mWARNING\033[0m:'



### Classes ###

class Expression:
    '''
    Container class for KLL expressions
    '''

    def __init__(self, lparam, operator, rparam, context):
        '''
        Initialize expression container

        @param lparam:   LOperatorData token
        @param operator: Operator token
        @param rparam:   ROperatorData token
        @param context:  Parent context of expression
        '''
        # First stage/init
        self.lparam_token = lparam
        self.operator_token = operator
        self.rparam_token = rparam
        self.context = context  # TODO, set multiple contexts for later stages

        # Second stage
        self.lparam_sub_tokens = []
        self.rparam_sub_tokens = []

        # BaseMap expression
        self.base_map = False

        # Default ConnectId
        self.connect_id = 0

        # Mutate class into the desired type
        self.__class__ = {
            '=>': NameAssociationExpression,
            '<=': DataAssociationExpression,
            '=': AssignmentExpression,
            ':': MapExpression,
        }[self.operator_type()]

    def operator_type(self):
        '''
        Determine which base operator this operator is of

        All : (map) expressions are tokenized/parsed the same way

        @return Base string representation of the operator
        '''
        if ':' in self.operator_token.value:
            return ':'

        return self.operator_token.value

    def final_tokens(self, no_filter=False):
        '''
        Return the final list of tokens, must complete the second stage first

        @param no_filter: If true, do not filter out Space tokens

        @return Finalized list of tokens
        '''
        ret = self.lparam_sub_tokens + \
            [self.operator_token] + self.rparam_sub_tokens

        if not no_filter:
            ret = [x for x in ret if x.type != 'Space']
        return ret

    def regen_str(self):
        '''
        Re-construct the string based off the original set of tokens

        <lparam><operator><rparam>;
        '''
        return "{0}{1}{2};".format(
            self.lparam_token.value,
            self.operator_token.value,
            self.rparam_token.value,
        )

    def point_chars(self, pos_list):
        '''
        Using the regenerated string, point to a given list of characters
        Used to indicate where a possible issue/syntax error is

        @param pos_list: List of character indices

        i.e.
        > U"A" : : U"1";
        >        ^
        '''
        out = "\t{0}\n\t".format(self.regen_str())

        # Place a ^ character at the given locations
        curpos = 1
        for pos in sorted(pos_list):
            # Pad spaces, then add a ^
            out += ' ' * (pos - curpos)
            out += '^'
            curpos += pos

        return out

    def rparam_start(self):
        '''
        Starting positing char of rparam_token in a regen_str
        '''
        return len(self.lparam_token.value) + len(self.operator_token.value)

    def __repr__(self):
        # Build string representation based off of what has been set
        # lparam, operator and rparam are always set
        out = "Expression: {0}{1}{2}".format(
            self.lparam_token.value,
            self.operator_token.value,
            self.rparam_token.value,
        )

        # TODO - Add more depending on what has been set
        return out

    def kllify(self):
        '''
        Returns KLL version of the expression

        May not look like the original expression if simplication has taken place
        '''
        print(
            "{0} kllify not defined for {1}".format(
                WARNING,
                self.__class__.__name__))
        out = "{0}{1}{2};".format(
            self.lparam_token.value,
            self.operator_token.value,
            self.rparam_token.value,
        )
        return out

    def unique_keys(self):
        '''
        Generates a list of unique identifiers for the expression that is mergeable
        with other functional equivalent expressions.

        This method should never get called directly as a generic Expression
        '''
        return [('UNKNOWN KEY', 'UNKNOWN EXPRESSION')]


class AssignmentExpression(Expression):
    '''
    Container class for assignment KLL expressions
    '''
    type = None
    name = None
    pos = None
    value = None

    ## Setters ##
    def array(self, name, pos, value):
        '''
        Assign array assignment parameters to expression

        @param name:  Name of variable
        @param pos:   Array position of the value (if None, overwrite the entire array)
        @param value: Value of the array, if pos is specified, this is the value of an element

        @return: True if parsing was successful
        '''
        self.type = 'Array'
        self.name = name
        self.pos = pos
        self.value = value

        # If pos is not none, flatten
        if pos is not None:
            self.value = "".join(str(x) for x in self.value)

        return True

    def merge_array(self, new_expression=None):
        '''
        Merge arrays, used for position assignments
        Merges unconditionally, make sure this is what you want to do first

        If no additional array is specified, just "cap-off" array.
        This does a proper array expansion into a python list.

        @param new_expression: AssignmentExpression type array, ignore if None
        '''
        # First, check if base expression needs to be capped
        if self.pos is not None:
            # Generate a new string array
            new_value = [""] * self.pos

            # Append the old contents to the list
            new_value.append(self.value)
            self.value = new_value

            # Clear pos, to indicate that array has been capped
            self.pos = None

        # Next, if a new_expression has been specified, merge in
        if new_expression is not None and new_expression.pos is not None:
            # Check if we need to extend the list
            new_size = new_expression.pos + 1 - len(self.value)
            if new_size > 0:
                self.value.extend([""] * new_size)

            # Assign value to array
            self.value[new_expression.pos] = new_expression.value

    def variable(self, name, value):
        '''
        Assign variable assignment parameters to expression

        @param name: Name of variable
        @param value: Value of variable

        @return: True if parsing was successful
        '''
        self.type = 'Variable'
        self.name = name
        self.value = value

        # Flatten value, often a list of various token types
        self.value = "".join(str(x) for x in self.value)

        return True

    def __repr__(self):
        if self.type == 'Variable':
            return "{0} = {1};".format(self.name, self.value)
        elif self.type == 'Array':
            # Output KLL style array, double quoted elements, space-separated
            if isinstance(self.value, list):
                output = "{0}[] =".format(self.name)
                for value in self.value:
                    output += ' "{0}"'.format(value)
                output += ";"
                return output

            # Single array assignment
            else:
                return "{0}[{1}] = {2};".format(
                    self.name, self.pos, self.value)

        return "ASSIGNMENT UNKNOWN"

    def kllify(self):
        '''
        Returns KLL version of the expression

        May not look like the original expression if simplication has taken place

        __repr__ is formatted correctly with assignment expressions
        '''
        return self.__repr__()

    def unique_keys(self):
        '''
        Generates a list of unique identifiers for the expression that is mergeable
        with other functional equivalent expressions.
        '''
        return [(self.name, self)]


class NameAssociationExpression(Expression):
    '''
    Container class for name association KLL expressions
    '''
    type = None
    name = None
    association = None

    ## Setters ##
    def capability(self, name, association, parameters):
        '''
        Assign a capability C function name association

        @param name: Name of capability
        @param association: Name of capability in target backend output

        @return: True if parsing was successful
        '''
        self.type = 'Capability'
        self.name = name
        self.association = CapId(association, 'Definition', parameters)

        return True

    def define(self, name, association):
        '''
        Assign a define C define name association

        @param name: Name of variable
        @param association: Name of association in target backend output

        @return: True if parsing was successful
        '''
        self.type = 'Define'
        self.name = name
        self.association = association

        return True

    def __repr__(self):
        return "{0} <= {1};".format(self.name, self.association)

    def kllify(self):
        '''
        Returns KLL version of the expression

        May not look like the original expression if simplication has taken place
        '''
        return "{0}".format(self)

    def unique_keys(self):
        '''
        Generates a list of unique identifiers for the expression that is mergeable
        with other functional equivalent expressions.
        '''
        return [(self.name, self)]


class DataAssociationExpression(Expression):
    '''
    Container class for data association KLL expressions
    '''
    type = None
    association = None
    value = None

    ## Setters ##
    def animation(self, animations, animation_modifiers):
        '''
        Animation definition and configuration

        @return: True if parsing was successful
        '''
        self.type = 'Animation'
        self.association = animations
        self.value = animation_modifiers

        return True

    def animationFrame(self, animation_frames, pixel_modifiers):
        '''
        Pixel composition of an Animation Frame

        @return: True if parsing was successful
        '''

        self.type = 'AnimationFrame'
        self.association = animation_frames
        self.value = pixel_modifiers

        return True

    def pixelPosition(self, pixels, position):
        '''
        Pixel Positioning

        @return: True if parsing was successful
        '''
        for pixel in pixels:
            pixel.setPosition(position)

        self.type = 'PixelPosition'
        self.association = pixels

        return True

    def scanCodePosition(self, scancodes, position):
        '''
        Scan Code to Position Mapping

        Note: Accepts lists of scan codes
        Alone this isn't useful, but you can assign rows and columns using ranges instead of individually

        @return: True if parsing was successful
        '''
        for scancode in scancodes:
            scancode.setPosition(position)

        self.type = 'ScanCodePosition'
        self.association = scancodes

        return True

    def update(self, new_expression):
        '''
        Update expression

        @param new_expression: Expression used to update this one
        '''
        supported = ['PixelPosition', 'ScanCodePosition']
        if new_expression.type in supported:
            for scancode in self.association:
                scancode.updatePositions(new_expression.association[0])

    def __repr__(self):
        if self.type in ['PixelPosition', 'ScanCodePosition']:
            output = ""
            for index, association in enumerate(self.association):
                if index > 0:
                    output += "; "
                output += "{0}".format(association)
            return "{0};".format(output)
        return "{0} <= {1};".format(self.association, self.value)

    def kllify(self):
        '''
        Returns KLL version of the expression

        May not look like the original expression if simplication has taken place

        __repr__ is formatted correctly with assignment expressions
        '''

        if self.type in ['PixelPosition', 'ScanCodePosition']:
            output = ""
            for index, association in enumerate(self.association):
                if index > 0:
                    output += "; "
                output += "{0}".format(association.kllify())
            return "{0};".format(output)

        if self.type in ['AnimationFrame']:
            output = "{0} <= ".format(self.association[0].kllify())
            for index, association in enumerate(self.value):
                if index > 0:
                    output += ", "
                output += "{0}".format(association[0].kllify())
            return "{0};".format(output)

        return "{0} <= {1};".format(
            self.association.kllify(), self.value.kllify())

    def unique_keys(self):
        '''
        Generates a list of unique identifiers for the expression that is mergeable
        with other functional equivalent expressions.
        '''
        keys = []

        # Positions require a bit more introspection to get the unique keys
        if self.type in ['PixelPosition', 'ScanCodePosition']:
            for index, key in enumerate(self.association):
                uniq_expr = self

                # If there is more than one key, copy the expression
                # and remove the non-related variants
                if len(self.association) > 1:
                    uniq_expr = copy.copy(self)

                    # Isolate variant by index
                    uniq_expr.association = [uniq_expr.association[index]]

                keys.append(("{0}".format(key.unique_key()), uniq_expr))

        # AnimationFrames are already list of keys
        # TODO Reorder frame assignments to dedup function equivalent mappings
        elif self.type in ['AnimationFrame']:
            for index, key in enumerate(self.association):
                uniq_expr = self

                # If there is more than one key, copy the expression
                # and remove the non-related variants
                if len(self.association) > 1:
                    uniq_expr = copy.copy(self)

                    # Isolate variant by index
                    uniq_expr.association = [uniq_expr.association[index]]

                keys.append(("{0}".format(key), uniq_expr))

        # Otherwise treat as a single element
        else:
            keys = [("{0}".format(self.association), self)]

        # Remove any duplicate keys
        # TODO Stat? Might be at neat report about how many duplicates were
        # squashed
        keys = list(set(keys))

        return keys


class MapExpression(Expression):
    '''
    Container class for KLL map expressions
    '''
    type = None
    triggers = None
    operator = None
    results = None
    animation = None
    animation_frame = None
    pixels = None
    position = None
    trigger_identifiers = ['IndCode', 'GenericTrigger', 'Layer', 'LayerLock', 'LayerShift', 'LayerLatch', 'ScanCode']

    def __init__(self, triggers, operator, results):
        '''
        Initialize MapExpression

        Used when copying MapExpressions from different expressions

        @param triggers: Sequence of combos of ranges of namedtuples
        @param operator: Type of map operation
        @param results: Sequence of combos of ranges of namedtuples
        '''
        self.type = 'TriggerCode'
        self.triggers = triggers
        self.operator = operator
        self.results = results

        self.connect_id = 0

    ## Setters ##
    def triggerCode(self, triggers, operator, results):
        '''
        Trigger Code mapping
        Takes in any combination of triggers and sets the expression accordingly.

        @param triggers: Sequence of combos of ranges of namedtuples
        @param operator: Type of map operation
        @param results: Sequence of combos of ranges of namedtuples

        @return: True if parsing was successful
        '''
        self.type = 'TriggerCode'
        self.triggers = triggers
        self.operator = operator
        self.results = results

        return True

    def pixelChannels(self, pixelmap, trigger):
        '''
        Pixel Channel Composition

        @return: True if parsing was successful
        '''
        self.type = 'PixelChannel'
        self.pixel = pixelmap
        self.position = trigger

        return True

    def triggersSequenceOfCombosOfIds(self, index=0):
        '''
        Takes triggers and converts into explicit ids

        Only uses the first index by default.
        @param index: Which trigger sequence to expand
        @return: list of lists

        Example (index=0)
        [[[S10, S16], [S42]], [[S11, S16], [S42]]] -> [[10, 16], [42]]
        '''
        nsequence = []
        for combo in self.triggers[index]:
            ncombo = []
            for identifier in combo:
                ncombo.append(identifier.json())
            nsequence.append(ncombo)
        return nsequence

    def resultsSequenceOfCombosOfIds(self, index=0):
        '''
        Takes results and converts into explicit capabilities

        Only uses the first index by default.
        @param index: Which result sequence to expand
        @return: list of lists
        '''
        nsequence = []
        for combo in self.results[index]:
            ncombo = []
            for identifier in combo:
                ncombo.append(identifier.json())
            nsequence.append(ncombo)
        return nsequence

    def sequencesOfCombosOfIds(self, expression_param):
        '''
        Prettified Sequence of Combos of Identifiers

        @param expression_param: Trigger or Result parameter of an expression

        Scan Code Example
        [[[S10, S16], [S42]], [[S11, S16], [S42]]] -> (S10 + S16, S42)|(S11 + S16, S42)
        '''
        output = ""

        # Sometimes during error cases, might be None
        if expression_param is None:
            return output

        # Iterate over each trigger/result variants (expanded from ranges),
        # each one is a sequence
        for index, sequence in enumerate(expression_param):
            if index > 0:
                output += "|"
            output += "("

            # Iterate over each combo (element of the sequence)
            for index, combo in enumerate(sequence):
                if index > 0:
                    output += ", "

                # Iterate over each trigger identifier
                for index, identifier in enumerate(combo):
                    if index > 0:
                        output += " + "
                    output += "{0}".format(identifier)

            output += ")"

        return output

    def sequencesOfCombosOfIds_kll(self, expression_param):
        '''
        Prettified Sequence of Combos of Identifiers, kll output edition

        @param expression_param: Trigger or Result parameter of an expression

        Scan Code Example
        [[[S10, S16], [S42]], [[S11, S16], [S42]]] -> ['S10 + S16, S42', 'S11 + S16, S42']
        '''
        output = ['']

        # Sometimes during error cases, might be None
        if expression_param is None:
            return output

        # Iterate over each trigger/result variants (expanded from ranges),
        # each one is a sequence
        for index, sequence in enumerate(expression_param):
            if index > 0:
                output.append('')

            # Iterate over each combo (element of the sequence)
            for index, combo in enumerate(sequence):
                if index > 0:
                    output[-1] += ", "

                # Iterate over each trigger identifier
                for index, identifier in enumerate(combo):
                    if index > 0:
                        output[-1] += " + "
                    output[-1] += "{0}".format(identifier.kllify())

        return output

    def trigger_id_list(self):
        '''
        Returns a list of ids within the sequence of combos
        May contain duplicates
        '''
        id_list = []

        # Iterate over each trigger/result variants (expanded from ranges)
        for sequence in self.triggers:
            # Iterate over each combo (element of the sequence)
            for combo in sequence:
                # Iterate over each trigger identifier
                for identifier in combo:
                    id_list.append(identifier)

        return id_list

    def min_trigger_uid(self):
        '''
        Returns the min numerical uid
        Used for trigger identifiers
        '''
        min_uid = 0xFFFF

        # Iterate over list of identifiers in trigger
        for identifier in self.trigger_id_list():
            if identifier.type in self.trigger_identifiers and identifier.get_uid() < min_uid:
                min_uid = identifier.get_uid()

        return min_uid

    def max_trigger_uid(self):
        '''
        Returns the max numerical uid
        Used for trigger identifiers
        '''
        max_uid = 0

        # Iterate over list of identifiers in trigger
        for identifier in self.trigger_id_list():
            if identifier.type in self.trigger_identifiers and identifier.get_uid() > max_uid:
                max_uid = identifier.get_uid()

        return max_uid

    def add_trigger_uid_offset(self, offset):
        '''
        Adds a uid/scancode offset to all triggers

        This is used when applying the connect_id interconnect offset during mapping indices generation
        '''
        # Iterate over list of identifiers in trigger
        for identifier in self.trigger_id_list():
            if identifier.type == 'ScanCode':
                identifier.updated_uid = identifier.uid + offset

    def elems(self):
        '''
        Return number of trigger and result elements

        Useful for determining if this is a trigger macro (2+)
        Should always return at least (1,1) unless it's an invalid calculation

        @return: ( triggers, results )
        '''
        elems = [0, 0]

        # XXX Needed?
        if self.type == 'PixelChannel':
            return tuple(elems)

        # Iterate over each trigger variant (expanded from ranges), each one is
        # a sequence
        for sequence in self.triggers:
            # Iterate over each combo (element of the sequence)
            for combo in sequence:
                # Just measure the size of the combo
                elems[0] += len(combo)

        # Iterate over each result variant (expanded from ranges), each one is
        # a sequence
        for sequence in self.results:
            # Iterate over each combo (element of the sequence)
            for combo in sequence:
                # Just measure the size of the combo
                elems[1] += len(combo)

        return tuple(elems)

    def trigger_str(self):
        '''
        String version of the trigger
        Used for sorting
        '''
        # Pixel Channel Mapping doesn't follow the same pattern
        if self.type == 'PixelChannel':
            return "{0}".format(self.pixel)

        return "{0}".format(
            self.sequencesOfCombosOfIds(self.triggers),
        )

    def result_str(self):
        '''
        String version of the result
        Used for sorting
        '''
        # Pixel Channel Mapping doesn't follow the same pattern
        if self.type == 'PixelChannel':
            return "{0}".format(self.position)

        return "{0}".format(
            self.sequencesOfCombosOfIds(self.results),
        )

    def __repr__(self):
        # Pixel Channel Mapping doesn't follow the same pattern
        if self.type == 'PixelChannel':
            return "{0} : {1};".format(self.pixel, self.position)

        return "{0} {1} {2};".format(
            self.sequencesOfCombosOfIds(self.triggers),
            self.operator,
            self.sequencesOfCombosOfIds(self.results),
        )

    def sort_trigger(self):
        '''
        Returns sortable trigger
        '''
        if self.type == 'PixelChannel':
            return "{0}".format(self.pixel.kllify())

        return "{0}".format(
            self.sequencesOfCombosOfIds_kll(self.triggers)[0],
        )

    def sort_result(self):
        '''
        Returns sortable result
        '''
        if self.type == 'PixelChannel':
            result = self.position
            # Handle None pixel mapping case
            if isinstance(self.position, list):
                result = self.sequencesOfCombosOfIds_kll(self.position)[0]
            return "{0}".format(result)

        return "{0}".format(
            self.sequencesOfCombosOfIds_kll(self.results)[0],
        )

    def kllify(self):
        '''
        Returns KLL version of the expression

        May not look like the original expression if simplication has taken place
        '''
        # TODO Handle variations? Instead of just the first index

        if self.type == 'PixelChannel':
            result = self.position
            # Handle None pixel mapping case
            if isinstance(self.position, list):
                result = self.sequencesOfCombosOfIds_kll(self.position)[0]
            return "{0} : {1};".format(self.pixel.kllify(), result)

        return "{0} {1} {2};".format(
            self.sequencesOfCombosOfIds_kll(self.triggers)[0],
            self.operator,
            self.sequencesOfCombosOfIds_kll(self.results)[0],
        )

    def unique_keys(self):
        '''
        Generates a list of unique identifiers for the expression that is mergeable
        with other functional equivalent expressions.

        TODO: This function should re-order combinations to generate the key.
        The final generated combo will be in the original order.
        '''
        keys = []

        # Pixel Channel only has key per mapping
        if self.type == 'PixelChannel':
            keys = [("{0}".format(self.pixel), self)]

        # Split up each of the keys
        else:
            # Iterate over each trigger/result variants (expanded from ranges),
            # each one is a sequence
            for index, sequence in enumerate(self.triggers):
                key = ""
                uniq_expr = self

                # If there is more than one key, copy the expression
                # and remove the non-related variants
                if len(self.triggers) > 1:
                    uniq_expr = copy.copy(self)

                    # Isolate variant by index
                    uniq_expr.triggers = [uniq_expr.triggers[index]]

                # Iterate over each combo (element of the sequence)
                for index, combo in enumerate(sequence):
                    if index > 0:
                        key += ", "

                    # Iterate over each trigger identifier
                    for index, identifier in enumerate(combo):
                        if index > 0:
                            key += " + "
                        key += "{0} {1}".format(self.connect_id, identifier)

                # Add key to list
                keys.append((key, uniq_expr))

        # Remove any duplicate keys
        # TODO Stat? Might be at neat report about how many duplicates were
        # squashed
        keys = list(set(keys))

        return keys



#!/usr/bin/env python3
'''
KLL Parsing Expressions

This file contains various parsing rules and processors used by funcparserlib for KLL

REMEMBER: When editing parser BNF-like expressions, order matters. Specifically lexer tokens and parser |
'''

# Parser doesn't play nice with linters, disable some checks
# pylint: disable=no-self-argument, too-many-public-methods, no-self-use, bad-builtin

# Copyright (C) 2016-2018 by Jacob Alexander
#
# This file is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This file is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this file.  If not, see <http://www.gnu.org/licenses/>.

### Imports ###

from kll.common.id import (
    AnimationId, AnimationFrameId,
    CapArgId, CapArgValue, CapId,
    HIDId,
    LayerId,
    NoneId,
    PixelAddressId, PixelId, PixelLayerId,
    ScanCodeId,
    TriggerId
)
from kll.common.modifier import AnimationModifierList
from kll.common.schedule import AnalogScheduleParam, ScheduleParam, Time

from kll.extern.funcparserlib.lexer import Token
from kll.extern.funcparserlib.parser import (some, a, many, oneplus, skip, maybe)



### Decorators ###

# Print Decorator Variables
ERROR = '\033[5;1;31mERROR\033[0m:'
WARNING = '\033[5;1;33mWARNING\033[0m:'



### Classes ###

# Parsing Functions

class Make:
    '''
    Collection of parse string interpreters
    '''

    def scanCode(token):
        '''
        Converts a raw scan code string into an ScanCodeId /w integer

        S0x10 -> 16
        '''
        if isinstance(token, int):
            return ScanCodeId(token)
        else:
            return ScanCodeId(int(token[1:], 0))

    def hidCode(type, token):
        '''
        Convert a given raw hid token string to an integer /w a type

        U"Enter" -> USB, Enter(0x28)
        '''
        # If already converted to a HIDId, just return
        if isinstance(token, HIDId):
            return token

        # If first character is a U or I, strip
        token_val = token.value
        if token_val[0] == "U" or token_val[0] == "I":
            token_val = token_val[1:]
        # CONS specifier
        elif 'CONS' in token_val:
            token_val = token_val[4:]
        # SYS specifier
        elif 'SYS' in token_val:
            token_val = token_val[3:]

        # Determine locale
        locale = token.locale

        # Determine lookup dictionary
        lookup = None
        if type == 'USBCode':
            lookup = locale.dict('from_hid_keyboard', key_caps=True)
        elif type == 'SysCode':
            lookup = locale.dict('from_hid_sysctrl', key_caps=True)
        elif type == 'ConsCode':
            lookup = locale.dict('from_hid_consumer', key_caps=True)
        elif type == 'IndCode':
            lookup = locale.dict('from_hid_led', key_caps=True)

        # If using string representation of USB Code, do lookup, case-insensitive
        if '"' in token_val:
            try:
                match_name = token_val[1:-1].upper()
                hid_code = int(lookup[match_name], 0)
            except LookupError as err:
                print("{} {} ({}) is an invalid USB HID Code Lookup...".format(
                    ERROR,
                    err,
                    locale
                ))
                raise
        else:
            # Already tokenized
            if (
                type == 'USBCode' and token_val[0] == 'USB'
                    or
                type == 'SysCode' and token_val[0] == 'SYS'
                    or
                type == 'ConsCode' and token_val[0] == 'CONS'
                    or
                type == 'IndCode' and token_val[0] == 'IND'
            ):
                hid_code = token_val[1]
            # Convert
            else:
                hid_code = int(token_val, 0)

        return HIDId(type, hid_code, locale)

    def usbCode(token):
        '''
        Convert a given raw USB Keyboard hid token string to an integer /w a type

        U"Enter" -> USB, Enter(0x28)
        '''
        return Make.hidCode('USBCode', token)

    def consCode(token):
        '''
        Convert a given raw Consumer Control hid token string to an integer /w a type
        '''
        return Make.hidCode('ConsCode', token)

    def sysCode(token):
        '''
        Convert a given raw System Control hid token string to an integer /w a type
        '''
        return Make.hidCode('SysCode', token)

    def indCode(token):
        '''
        Convert a given raw Indicator hid token string to an integer /w a type
        '''
        return Make.hidCode('IndCode', token)

    def animation(name):
        '''
        Converts a raw animation value into an AnimationId /w name

        A"myname" -> myname
        '''
        if name[0] == "A":
            return AnimationId(name[2:-1])
        else:
            return AnimationId(name)

    def animationTrigger(animation, specifier):
        '''
        Generate an AnimationId
        '''
        trigger_list = []

        # AnimationId
        trigger_list.append(AnimationId(animation))

        return trigger_list, specifier

    def animationAssociation(animation, frame_identifier):
        '''
        Generate an AnimationFrameId
        '''
        trigger_list = []

        # AnimationFrameId
        for index in frame_identifier:
            trigger_list.append([[AnimationFrameId(animation, index)]])

        return trigger_list

    def animationCapability(animation, modifiers):
        '''
        Apply modifiers to AnimationId
        '''
        if modifiers is not None:
            animation.setModifiers(modifiers)
        return [animation]

    def animationModlist(modifiers):
        '''
        Build an AnimationModifierList

        Only used for animation data association
        '''
        modlist = AnimationModifierList()
        modlist.setModifiers(modifiers)
        return modlist

    def pixelCapability(pixels, modifiers):
        '''
        Apply modifiers to list of pixels/pixellists

        Results in a combination of pixel capabilities
        '''
        pixelcap_list = []
        for pixel in pixels:
            pixel.setModifiers(modifiers)
            pixelcap_list.append(pixel)
        return pixelcap_list

    def pixel(token):
        '''
        Converts a raw pixel value into a PixelId /w integer

        P0x3 -> 3
        '''
        if isinstance(token, int):
            return PixelId(token)
        else:
            return PixelId(int(token[1:], 0))

    def pixel_list(pixel_list):
        '''
        Converts a list a numbers into a list of PixelIds
        '''
        pixels = []
        for pixel in pixel_list:
            pixels.append(PixelId(pixel))
        return pixels

    def pixelLayer(token):
        '''
        Converts a raw pixel layer value into a PixelLayerId /w integer

        PL0x3 -> 3
        '''
        if isinstance(token, int):
            return PixelLayerId(token)
        else:
            return PixelLayerId(int(token[2:], 0))

    def pixelLayer_list(layer_list):
        '''
        Converts a list a numbers into a list of PixelLayerIds
        '''
        layers = []
        for layer in layer_list:
            layers.append(PixelLayerId(layer))
        return layers

    def pixelchan(pixel_list, chans):
        '''
        Apply channels to PixelId

        Only one pixel at a time can be mapped, hence pixel_list[0]
        '''
        pixel = pixel_list[0]
        pixel.setChannels(chans)
        return pixel

    def pixelmod(pixels, modifiers):
        '''
        Apply modifiers to list of pixels/pixellists

        Results in a combination of pixel capabilities
        '''
        pixelcap_list = []
        for pixel in pixels:
            # Convert HIDIds into PixelIds
            if isinstance(pixel, HIDId) or isinstance(pixel, ScanCodeId):
                pixel = PixelId(pixel)
            pixel.setModifiers(modifiers)
            pixelcap_list.append(pixel)
        return pixelcap_list

    def pixel_address(elems):
        '''
        Parse pixel positioning for row/column addressing

        @param elems:  index list or (operator, value)

        40
        c:0
        c:30%
        r:i+30
        '''
        pixel_address_list = []

        # Index list
        if isinstance(elems, list):
            # List of integers, possibly a range
            if isinstance(elems[0], int):
                for elem in elems:
                    pixel_address_list.append(PixelAddressId(index=elem))
            # Already ready to append
            elif isinstance(elems[0], PixelId):
                pixel_address_list.append(elems[0])

        # No value
        elif isinstance(elems, Token):
            # Row
            if "r:i" in elems.name:
                pixel_address_list.append(PixelAddressId(relRow=0))
            # Column
            if "c:i" in elems.name:
                pixel_address_list.append(PixelAddressId(relCol=0))

        # Operator with value
        elif isinstance(elems[0], Token):
            # Prepare address value
            value = elems[1]

            # Positioning
            if elems[0].type == "ColRowOperator":
                # Row
                if elems[0].name == "r:":
                    pixel_address_list.append(PixelAddressId(row=value))
                # Column
                if elems[0].name == "c:":
                    pixel_address_list.append(PixelAddressId(col=value))

            # Relative Positioning
            elif elems[0].type == "RelCROperator":
                if '-' in elems[0].name:
                    value *= -1

                # Row
                if "r:i" in elems[0].name:
                    pixel_address_list.append(PixelAddressId(relRow=value))
                # Column
                if "c:i" in elems[0].name:
                    pixel_address_list.append(PixelAddressId(relCol=value))

        return pixel_address_list

    def pixel_address_merge(elems):
        '''
        Merge pixel addresses together
        '''
        # Merge is only necessary if there is more than one element
        if len(elems) > 1:
            for elem in elems[1:]:
                elems[0].merge(elem)

        return [elems[0]]

    def position(token):
        '''
        Physical position split

        x:20 -> (x, 20)
        '''
        return token.split(':')

    def usbCode_number(token):
        '''
        USB Keyboard HID Code lookup
        '''
        return HIDId('USBCode', token.value, token.locale)

    def consCode_number(token):
        '''
        Consumer Control HID Code lookup
        '''
        return HIDId('ConsCode', token.value, token.locale)

    def sysCode_number(token):
        '''
        System Control HID Code lookup
        '''
        return HIDId('SysCode', token.value, token.locale)

    def indCode_number(token):
        '''
        Indicator HID Code lookup
        '''
        return HIDId('IndCode', token.value, token.locale)

    def none(token):
        '''
        Replace key-word with NoneId specifier (which indicates a noneOut capability)
        '''
        return [[[NoneId()]]]

    def seqString(token, spec='lspec'):
        '''
        Converts sequence string to a sequence of combinations

        'Ab'  -> U"Shift" + U"A", U"B"
        'abb' -> U"A", U"B", U"NoEvent", U"B"

        @param spec: 'lspec' or 'rspec'
        '''
        # Determine locale
        locale = token.locale

        # Compose string using set locale
        sequence = None
        if spec == 'lspec':
            sequence = locale.compose(token.value[1:-1], minimal_clears=True, no_clears=True)
        else:
            sequence = locale.compose(token.value[1:-1], minimal_clears=True)

        # Convert each element in sequence of combos to HIDIds
        hid_ids = []
        for combo in sequence:
            new_combo = []
            for elem in combo:
                # Lookup uid (usb code) from alias name (used in sequence)
                new_elem = HIDId('USBCode', int(locale.json()['from_hid_keyboard'][elem], 0), locale)
                new_combo.append(new_elem)
            hid_ids.append(new_combo)

        return hid_ids

    def seqStringL(token):
        '''
        Converts sequence string to a sequence of combinations
        lspec side

        'Ab'  -> U"Shift" + U"A", U"B"
        'abb' -> U"A", U"B", U"NoEvent", U"B"
        '''
        return Make.seqString(token, 'lspec')

    def seqStringR(token):
        '''
        Converts sequence string to a sequence of combinations
        rspec side

        'Ab'  -> U"Shift" + U"A", U"B"
        'abb' -> U"A", U"B", U"NoEvent", U"B"
        '''
        return Make.seqString(token, 'rspec')

    def string(token):
        '''
        Converts a raw string to a Python string

        "this string" -> this string
        '''
        return token[1:-1]

    def unseqString(token):
        '''
        Converts a raw sequence string to a Python string

        'this string' -> this string
        '''
        return token[1:-1]

    def number(token):
        '''
        Convert string number to Python integer
        '''
        return int(token, 0)

    def numberToken(token):
        '''
        Convert token value to Python integer
        '''
        try:
            token.value = int(token.value, 0)
        except TypeError:
            pass
        return token

    def percent(token):
        '''
        Convert string percent to Python float
        '''
        return int(token[:-1], 0) / 100.0

    def timing(token):
        '''
        Convert raw timing parameter to integer time and determine units

        1ms -> 1, ms
        '''
        # Find ms, us, or s
        if 'ms' in token:
            unit = 'ms'
            num = token.split('m')[0]
        elif 'us' in token:
            unit = 'us'
            num = token.split('u')[0]
        elif 'ns' in token:
            unit = 'ns'
            num = token.split('n')[0]
        elif 's' in token:
            unit = 's'
            num = token.split('s')[0]
        else:
            print("{0} cannot find timing unit in token '{1}'".format(ERROR, token))

        return Time(float(num), unit)

    def specifierTiming(timing):
        '''
        When only timing is given, infer state at a later stage from the context of the mapping
        '''
        return ScheduleParam(None, timing)

    def specifierState(state, timing=None):
        '''
        Generate a Schedule Parameter
        Automatically mutates itself into the correct object type
        '''
        return ScheduleParam(state, timing)

    def specifierAnalog(value):
        '''
        Generate an Analog Schedule Parameter
        '''
        return AnalogScheduleParam(value)

    def specifierUnroll(identifier, schedule_params):
        '''
        Unroll specifiers into the trigger/result identifier

        First, combine all Schedule Parameters into a Schedul
        Then attach Schedule to the identifier

        If the identifier is a list, then iterate through them
        and apply the schedule to each
        '''
        # Check if this is a list of identifiers
        if isinstance(identifier, list):
            for ident in identifier:
                ident.setSchedule(schedule_params)
            return identifier
        else:
            identifier.setSchedule(schedule_params)

        return [identifier]

    def layerTypeIdent(layer_type, inner_list, specifier):
        '''
        Given a layer expression, determine what kind of layer expression

        Layer
        LayerShift
        LayerLatch
        LayerLock
        '''
        # Determine layer type (remove [)
        layer_type = layer_type[:-1]

        # Add layer type to each given layer
        identifier_list = []
        for layer in inner_list:
            identifier_list.append(LayerId(layer_type, layer))

        return identifier_list, specifier

    def genericTriggerIdent(identifier, code, specifier):
        '''
        Given a generic trigger, create a TriggerId object

        Generic Triggers don't support ranges
        '''
        trigger_obj = TriggerId(identifier, code)

        return trigger_obj, specifier

    # Range can go from high to low or low to high
    def scanCode_range(rangeVals):
        '''
        Scan Code range expansion

        S[0x10-0x12] -> S0x10, S0x11, S0x12
        '''
        start = rangeVals[0]
        end = rangeVals[1]

        # Swap start, end if start is greater than end
        if start > end:
            start, end = end, start

        # Iterate from start to end, and generate the range
        values = list(range(start, end + 1))

        # Generate ScanCodeIds
        return [ScanCodeId(v) for v in values]

    # Range can go from high to low or low to high
    # Warn on 0-9 for USBCodes (as this does not do what one would expect) TODO
    # Lookup USB HID tags and convert to a number
    def hidCode_range(type, rangeVals):
        '''
        HID Code range expansion

        U["A"-"C"] -> U"A", U"B", U"C"
        '''

        # Check if already integers
        if isinstance(rangeVals[0], int):
            start = rangeVals[0]
        else:
            start = Make.hidCode(type, rangeVals[0]).uid

        if isinstance(rangeVals[1], int):
            end = rangeVals[1]
        else:
            end = Make.hidCode(type, rangeVals[1]).uid

        # Swap start, end if start is greater than end
        if start > end:
            start, end = end, start

        # Iterate from start to end, and generate the range
        listRange = list(range(start, end + 1))

        # Determine locale
        locale = rangeVals[0].locale

        # Convert each item in the list to a tuple
        for item in range(len(listRange)):
            listRange[item] = HIDId(type, listRange[item], locale)
        return listRange

    def usbCode_range(rangeVals):
        '''
        USB Keyboard HID Code range expansion
        '''
        return Make.hidCode_range('USBCode', rangeVals)

    def sysCode_range(rangeVals):
        '''
        System Control HID Code range expansion
        '''
        return Make.hidCode_range('SysCode', rangeVals)

    def consCode_range(rangeVals):
        '''
        Consumer Control HID Code range expansion
        '''
        return Make.hidCode_range('ConsCode', rangeVals)

    def indCode_range(rangeVals):
        '''
        Indicator HID Code range expansion
        '''
        return Make.hidCode_range('IndCode', rangeVals)

    def range(start, end):
        '''
        Converts a start and end points of a range to a list of numbers

        Can go low to high or high to low
        '''
        # High to low
        if end < start:
            return list(range(end, start + 1))

        # Low to high
        return list(range(start, end + 1))

    def capArg(argument, width=None):
        '''
        Converts a capability argument:width to a CapArgId

        If no width is specified, it is ignored
        '''
        return CapArgId(argument, width)

    def capArgValue(tuple_value):
        '''
        Converts a capability argument value to a CapArgValue
        '''
        sign, value = tuple_value
        if sign is not None:
            value *= -1
        return CapArgValue(value)

    def capUsage(name, arguments):
        '''
        Converts a capability tuple, argument list to a CapId Usage
        '''
        return CapId(name, 'Capability', arguments)

    def debug(tokens):
        '''
        Just prints tokens
        Used for debugging
        '''
        print(tokens)
        return tokens


### Rules ###

# Base Rules


def const(x): return lambda _: x


def unarg(f): return lambda x: f(*x)


def flatten(list): return sum(list, [])


def tokenValue(x):
    '''
    Return string value of a token

    @param x: Token
    @returns: String value of token
    '''
    return x.value


def tokenType(t):
    '''
    Returns string of token

    @param t: Name of token type
    @returns: String of token
    '''
    return some(lambda x: x.type == t) >> tokenValue


def tokenTypeOnly(t):
    '''
    Returns the full token object

    @param t: Name of token type
    @return: Token matching
    '''
    return some(lambda x: x.type == t)


def operator(s): return a(Token('Operator', s)) >> tokenValue


def parenthesis(s): return a(Token('Parenthesis', s)) >> tokenValue


def bracket(s): return a(Token('Bracket', s)) >> tokenValue


eol = a(Token('EndOfLine', ';'))


def maybeFlatten(items):
    '''
    Iterate through top-level lists
    Flatten, only if the element is also a list

    [[1,2],3,[[4,5]]] -> [1,2,3,[4,5]]
    '''
    new_list = []
    for elem in items:
        # Flatten only if a list
        if isinstance(elem, list):
            new_list.extend(elem)
        else:
            new_list.append(elem)
    return new_list


def listElem(item):
    '''
    Convert to a list element
    '''
    return [item]


def listToTuple(items):
    '''
    Convert list to a tuple
    '''
    return tuple(items)


def oneLayerFlatten(items):
    '''
    Flatten only the top layer (list of lists of ...)
    '''
    mainList = []
    for sublist in items:
        for item in sublist:
            mainList.append(item)

    return mainList


def optionCompression(sequence):
    '''
    Adds another dimension to a list of lists.

    This is the inverse operation of optionExpansion, iff there were no expanded ranges

    @param: sequence: Sequence of combos
    @returns: Squence of combos of ranges
    '''
    new_list = []
    for combo in sequence:
        new_combo = []
        for elem in combo:
            new_combo.append([elem])
        new_list.append(new_combo)

    return new_list


def optionExpansion(sequences):
    '''
    Expand ranges of values in the 3rd dimension of the list, to a list of 2nd lists

    i.e. [ sequence, [ combo, [ range ] ] ] --> [ [ sequence, [ combo ] ], <option 2>, <option 3> ]

    @param sequences: Sequence of combos of ranges
    @returns: List of sequences of combos
    '''
    expandedSequences = []

    # Total number of combinations of the sequence of combos that needs to be generated
    totalCombinations = 1

    # List of leaf lists, with number of leaves
    maxLeafList = []

    # Traverse to the leaf nodes, and count the items in each leaf list
    for sequence in sequences:
        for combo in sequence:
            rangeLen = len(combo)
            totalCombinations *= rangeLen
            maxLeafList.append(rangeLen)

    # Counter list to keep track of which combination is being generated
    curLeafList = [0] * len(maxLeafList)

    # Generate a list of permuations of the sequence of combos
    for count in range(0, totalCombinations):
        expandedSequences.append([])  # Prepare list for adding the new combination
        pos = 0

        # Traverse sequence of combos to generate permuation
        for sequence in sequences:
            expandedSequences[-1].append([])
            for combo in sequence:
                expandedSequences[-1][-1].append(combo[curLeafList[pos]])
                pos += 1

        # Increment combination tracker
        for leaf in range(0, len(curLeafList)):
            curLeafList[leaf] += 1

            # Reset this position, increment next position (if it exists), then stop
            if curLeafList[leaf] >= maxLeafList[leaf]:
                curLeafList[leaf] = 0
                if leaf + 1 < len(curLeafList):
                    curLeafList[leaf + 1] += 1

    return expandedSequences


def listit(t):
    '''
    Convert tuple of tuples to list of lists
    '''
    return list(map(listit, t)) if isinstance(t, (list, tuple)) else t


def tupleit(t):
    '''
    Convert list of lists to tuple of tuples
    '''
    return tuple(map(tupleit, t)) if isinstance(t, (tuple, list)) else t


# Sub Rules

usbCode = tokenTypeOnly('USBCode') >> Make.usbCode
scanCode = tokenType('ScanCode') >> Make.scanCode
consCode = tokenTypeOnly('ConsCode') >> Make.consCode
sysCode = tokenTypeOnly('SysCode') >> Make.sysCode
indCode = tokenTypeOnly('IndCode') >> Make.indCode
animation = tokenType('Animation') >> Make.animation
pixel = tokenType('Pixel') >> Make.pixel
pixelLayer = tokenType('PixelLayer') >> Make.pixelLayer
none = tokenType('None') >> Make.none
position = tokenType('Position') >> Make.position

comma = tokenType('Comma')
content = tokenType('VariableContents')
dash = tokenType('Dash')
name = tokenType('Name')
number = tokenType('Number') >> Make.number
numberToken = tokenTypeOnly('Number') >> Make.numberToken
percent = tokenType('Percent') >> Make.percent
plus = tokenType('Plus')
timing = tokenType('Timing') >> Make.timing

string = tokenType('String') >> Make.string
unString = tokenTypeOnly('String')  # When the double quotes are still needed for internal processing
seqStringL = tokenTypeOnly('SequenceStringL') >> Make.seqStringL >> optionCompression # lspec
seqStringR = tokenTypeOnly('SequenceStringR') >> Make.seqStringR >> optionCompression # rspec
unseqString = tokenType('SequenceString') >> Make.unseqString  # For use with variables


def colRowOperator(s): return a(Token('ColRowOperator', s))


def relCROperator(s): return a(Token('RelCROperator', s))


pixelOperator = tokenType('PixelOperator')

# Code variants
code_begin = tokenType('CodeBegin')
code_end = tokenType('CodeEnd')

# Specifier
specifier_basic = (timing >> Make.specifierTiming) | (name >> Make.specifierState)
specifier_complex = (name + skip(operator(':')) + timing) >> unarg(Make.specifierState)
specifier_state = specifier_complex | specifier_basic
specifier_analog = number >> Make.specifierAnalog
specifier_list = skip(parenthesis('(')) + many((specifier_state | specifier_analog) + skip(maybe(comma))) + skip(parenthesis(')'))

# Scan Codes
scanCode_start = tokenType('ScanCodeStart')
scanCode_range = number + skip(dash) + number >> Make.scanCode_range
scanCode_listElem = number >> Make.scanCode
scanCode_specifier = (scanCode_range | scanCode_listElem) + maybe(specifier_list) >> unarg(Make.specifierUnroll)
scanCode_innerList = many(scanCode_specifier + skip(maybe(comma))) >> flatten
scanCode_expanded = skip(scanCode_start) + scanCode_innerList + skip(code_end) + maybe(specifier_list) >> unarg(Make.specifierUnroll)
scanCode_elem = scanCode + maybe(specifier_list) >> unarg(Make.specifierUnroll)
scanCode_combo_elem = scanCode_expanded | scanCode_elem
scanCode_single = (skip(scanCode_start) + scanCode_listElem + skip(code_end)) | scanCode
scanCode_il_nospec = oneplus((scanCode_range | scanCode_listElem) + skip(maybe(comma)))
scanCode_nospecifier = skip(scanCode_start) + scanCode_il_nospec + skip(code_end)

# Cons Codes
consCode_start = tokenType('ConsCodeStart')
consCode_number = numberToken >> Make.consCode_number
consCode_range = (consCode_number | unString) + skip(dash) + (number | unString) >> Make.consCode_range
consCode_listElemTag = unString >> Make.consCode
consCode_listElem = (consCode_number | consCode_listElemTag)
consCode_specifier = (consCode_range | consCode_listElem) + maybe(specifier_list) >> unarg(Make.specifierUnroll)
consCode_innerList = oneplus(consCode_specifier + skip(maybe(comma))) >> flatten
consCode_expanded = skip(consCode_start) + consCode_innerList + skip(code_end) + maybe(specifier_list) >> unarg(Make.specifierUnroll)
consCode_elem = consCode + maybe(specifier_list) >> unarg(Make.specifierUnroll)
consCode_il_nospec = oneplus((consCode_range | consCode_listElem) + skip(maybe(comma)))
consCode_nospecifier = skip(consCode_start) + consCode_il_nospec + skip(code_end)

# Sys Codes
sysCode_start = tokenType('SysCodeStart')
sysCode_number = numberToken >> Make.sysCode_number
sysCode_range = (sysCode_number | unString) + skip(dash) + (number | unString) >> Make.sysCode_range
sysCode_listElemTag = unString >> Make.sysCode
sysCode_listElem = (sysCode_number | sysCode_listElemTag)
sysCode_specifier = (sysCode_range | sysCode_listElem) + maybe(specifier_list) >> unarg(Make.specifierUnroll)
sysCode_innerList = oneplus(sysCode_specifier + skip(maybe(comma))) >> flatten
sysCode_expanded = skip(sysCode_start) + sysCode_innerList + skip(code_end) + maybe(specifier_list) >> unarg(Make.specifierUnroll)
sysCode_elem = sysCode + maybe(specifier_list) >> unarg(Make.specifierUnroll)
sysCode_il_nospec = oneplus((sysCode_range | sysCode_listElem) + skip(maybe(comma)))
sysCode_nospecifier = skip(sysCode_start) + sysCode_il_nospec + skip(code_end)

# Indicator Codes
indCode_start = tokenType('IndicatorStart')
indCode_number = numberToken >> Make.indCode_number
indCode_range = (indCode_number | unString) + skip(dash) + (number | unString) >> Make.indCode_range
indCode_listElemTag = unString >> Make.indCode
indCode_listElem = (indCode_number | indCode_listElemTag)
indCode_specifier = (indCode_range | indCode_listElem) + maybe(specifier_list) >> unarg(Make.specifierUnroll)
indCode_innerList = oneplus(indCode_specifier + skip(maybe(comma))) >> flatten
indCode_expanded = skip(indCode_start) + indCode_innerList + skip(code_end) + maybe(specifier_list) >> unarg(Make.specifierUnroll)
indCode_elem = indCode + maybe(specifier_list) >> unarg(Make.specifierUnroll)
indCode_il_nospec = oneplus((indCode_range | indCode_listElem) + skip(maybe(comma)))
indCode_nospecifier = skip(indCode_start) + indCode_il_nospec + skip(code_end)

# USB Codes
usbCode_start = tokenType('USBCodeStart')
usbCode_number = numberToken >> Make.usbCode_number
usbCode_range = (usbCode_number | unString) + skip(dash) + (number | unString) >> Make.usbCode_range
usbCode_listElemTag = unString >> Make.usbCode
usbCode_listElem = (usbCode_number | usbCode_listElemTag)
usbCode_specifier = (usbCode_range | usbCode_listElem) + maybe(specifier_list) >> unarg(Make.specifierUnroll)
usbCode_il_nospec = oneplus((usbCode_range | usbCode_listElem) + skip(maybe(comma)))
usbCode_nospecifier = skip(usbCode_start) + usbCode_il_nospec + skip(code_end)
usbCode_innerList = oneplus(usbCode_specifier + skip(maybe(comma))) >> flatten
usbCode_expanded = skip(usbCode_start) + usbCode_innerList + skip(code_end) + maybe(specifier_list) >> unarg(Make.specifierUnroll)
usbCode_elem = usbCode + maybe(specifier_list) >> unarg(Make.specifierUnroll)

# HID Codes
hidCode_elem = usbCode_expanded | usbCode_elem | sysCode_expanded | sysCode_elem | consCode_expanded | consCode_elem | indCode_expanded | indCode_elem

# Layers
layer_start = tokenType('LayerStart')
layer_range = (number) + skip(dash) + (number) >> unarg(Make.range)
layer_listElem = number >> listElem
layer_innerList = oneplus((layer_range | layer_listElem) + skip(maybe(comma))) >> flatten
layer_expanded = layer_start + layer_innerList + skip(code_end) + maybe(specifier_list) >> unarg(Make.layerTypeIdent) >> unarg(Make.specifierUnroll)

# Generic Triggers
gtrigger_start = tokenType('TriggerStart')
gtrigger_parts = skip(gtrigger_start) + number + skip(comma) + number + skip(code_end) + maybe(specifier_list)
gtrigger_expanded = gtrigger_parts >> unarg(Make.genericTriggerIdent) >> unarg(Make.specifierUnroll)

# Pixels
pixel_start = tokenType('PixelStart')
pixel_range = (number) + skip(dash) + (number) >> unarg(Make.range) >> Make.pixel_address
pixel_listElem = number >> listElem >> Make.pixel_address
pixel_pos = (colRowOperator('c:') | colRowOperator('r:')) + (number | percent) >> Make.pixel_address
pixel_posRel = (relCROperator('c:i+') | relCROperator('c:i-') | relCROperator('r:i+') | relCROperator('r:i-')) + (number | percent) >> Make.pixel_address
pixel_posRelHere = (relCROperator('c:i') | relCROperator('r:i')) >> Make.pixel_address
pixel_posMerge = oneplus((pixel_pos | pixel_posRel | pixel_posRelHere) + skip(maybe(comma))) >> flatten >> Make.pixel_address_merge
pixel_innerList = ((oneplus((pixel_range | pixel_listElem | pixel_posMerge) + skip(maybe(comma))) >> flatten) | (pixel_posMerge)) >> Make.pixel_list
pixel_expanded = skip(pixel_start) + pixel_innerList + skip(code_end)
pixel_elem = pixel >> listElem >> Make.pixel_address

# Pixel Layer
pixellayer_start = tokenType('PixelLayerStart')
pixellayer_range = (number) + skip(dash) + (number) >> unarg(Make.range)
pixellayer_listElem = number >> listElem
pixellayer_innerList = oneplus((pixellayer_range | pixellayer_listElem) + skip(maybe(comma))) >> flatten >> Make.pixelLayer_list
pixellayer_expanded = skip(pixellayer_start) + pixellayer_innerList + skip(code_end)
pixellayer_elem = pixelLayer >> listElem

# Pixel Channels
pixelchan_chans = many(number + skip(operator(':')) + number + skip(maybe(comma)))
pixelchan_elem = ((pixel_expanded | pixel_elem) + skip(parenthesis('(')) + pixelchan_chans + skip(parenthesis(')'))) >> unarg(Make.pixelchan)

# HID Id for Pixel Mods
pixelmod_hid_elem = (usbCode | sysCode | consCode | indCode | scanCode) >> listElem
pixelmod_hid = pixelmod_hid_elem | usbCode_nospecifier | scanCode_nospecifier | consCode_nospecifier | sysCode_nospecifier | indCode_nospecifier

# Pixel Mods
pixelmod_modop = maybe(pixelOperator | plus | dash) >> listElem
pixelmod_modva = number >> listElem
pixelmod_mods = oneplus((pixelmod_modop + pixelmod_modva + skip(maybe(comma))) >> flatten)
pixelmod_layer = (pixellayer_expanded | pixellayer_elem)
pixelmod_index = (pixel_expanded | pixel_elem | pixelmod_hid | pixelmod_layer)
pixelmod_elem = pixelmod_index + skip(parenthesis('(')) + pixelmod_mods + skip(parenthesis(')')) >> unarg(Make.pixelmod)

# Pixel Capability
pixel_capability = pixelmod_elem

# Animations
animation_start = tokenType('AnimationStart')
animation_name = name
animation_frame_range = (number) + skip(dash) + (number) >> unarg(Make.range)
animation_name_frame = many((animation_frame_range | number) + skip(maybe(comma))) >> maybeFlatten
animation_def = skip(animation_start) + animation_name + skip(code_end) >> Make.animation
animation_expanded = skip(animation_start) + animation_name + skip(maybe(comma)) + animation_name_frame + skip(code_end) >> unarg(Make.animationAssociation)
animation_trigger = skip(animation_start) + animation_name + skip(code_end) + maybe(specifier_list) >> unarg(Make.animationTrigger) >> unarg(Make.specifierUnroll)
animation_flattened = animation_expanded >> flatten >> flatten
animation_elem = animation

# Animation Modifier
animation_modifier_arg = number | (name + skip(parenthesis('(')) + many(number + skip(maybe(comma))) + skip(parenthesis(')'))) | name
animation_modifier = many((name | number) + maybe(skip(operator(':')) + animation_modifier_arg) + skip(maybe(comma)))
animation_modlist = animation_modifier >> Make.animationModlist

# Animation Capability
animation_capability = ((animation_def | animation_elem) + maybe(skip(parenthesis('(')) + animation_modifier + skip(parenthesis(')')))) >> unarg(Make.animationCapability)

# Capabilities
capFunc_argument = (maybe(dash) + number) >> Make.capArgValue  # TODO Allow for symbolic arguments, i.e. arrays and variables
capFunc_arguments = many(capFunc_argument + skip(maybe(comma)))
capFunc_elem = name + skip(parenthesis('(')) + capFunc_arguments + skip(parenthesis(')')) >> unarg(Make.capUsage) >> listElem
capFunc_combo = oneplus((hidCode_elem | capFunc_elem | animation_capability | pixel_capability | layer_expanded) + skip(maybe(plus))) >> listElem
capFunc_sequence = oneplus((capFunc_combo | seqStringR) + skip(maybe(comma))) >> oneLayerFlatten

# Trigger / Result Codes
triggerCode_combo = oneplus((scanCode_combo_elem | hidCode_elem | layer_expanded | animation_trigger | gtrigger_expanded) + skip(maybe(plus))) >> listElem
triggerCode_sequence = oneplus((triggerCode_combo | seqStringL | seqStringR) + skip(maybe(comma))) >> oneLayerFlatten
triggerCode_outerList = triggerCode_sequence >> optionExpansion
resultCode_outerList = ((capFunc_sequence >> optionExpansion) | none)

# Positions
position_list = oneplus(position + skip(maybe(comma)))
